from pyg_base import getargspec, first, Dict, loop, zipper, as_list
from numba import njit

__all__ = ['compiled']

def compiled(function):
    res = njit(nogil = True)(function)
    res.fullargspec = getargspec(function)
    return res

#from pyg.base import passthru; compiled = passthru ## use this to get valid code-coverage for compiled functions

first_ = loop(dict, list)(first)


def _data_state(keys, values, output = 'data'):
    if isinstance(values, dict):
        return type(values)({k: _data_state(keys, v, output) for k, v in values.items()})
    elif isinstance(values, list):
        return type(values)([_data_state(keys, v, output) for v in values])
    output = as_list(output)
    assert keys[:len(output)] == output
    res = Dict(zip(output,values))
    if len(keys) > len(output):
        res['state'] = Dict(zipper(keys[len(output):], values[len(output):]))
    return res


def _ignore(ignore, data, state):
    state = {} if state is None else state
    if ignore is True:
        return None, {}
    elif ignore is False:
        return data, state
    ignore = as_list(ignore)
    if 'data' in ignore:
        data = None
    if 'state' in ignore:
        state = {}
    return data, state

# import pandas as pd
# from pyg.base import wrapper, is_ts, is_dict, is_tuple, getargspec, loop, first, Dict, zipper

# class persist_data(wrapper):
#     """
#     We work with state-persisting function that return a variable called 'data'
    
#     :Example:
#     -------
#     >>> from pyg import *
#     >>> ts = pd.Series(np.random.normal(0,1,(1000)), drange(-999))
#     >>> ts[ts<0.1] = np.nan

#     We can run ts in one go:
        
#     >>> both = ffill_(ts)

#     Or split it into two parts:

#     >>> old = ffill_(ts.iloc[:500])
    
#     old.data is the data for first 500 entries. We can use this data to speed up later calculations...
    
#     >>> new = ffill_(ts.iloc[500:], **(old-'data')) 
    
#     We note: 
#     1) the 'data' variable must be remove
#     2) We need to glue together old.data and new.data
    
#     However... persist_data does the work for you:
        
#     >>> persist_ffill_ = persist_data(ffill_)
#     >>> glued_with_old = persist_ffill_(ts, **old)  # Note that we pass on data and the full ts, but calculation is done only on the 'new' bit of ts
#     >>> glued_just_new = persist_ffill_(ts.iloc[500:], **old)  

#     >>> assert eq(glued_with_old, both)    
#     >>> assert eq(glued_just_new, both)    
    
#     """
#     def wrapped(self, *args, **kwargs):
#         data = kwargs.pop('data', None)
#         if data is not None and len(data)>0 and is_ts(data):
#             cutoff = data.index[-1]
#             args_ = [v[v.index>cutoff] if is_ts(v) and len(v) else v for v in args]
#             kwargs_ = {k : v[v.index>cutoff] if is_ts(v) and len(v) else v for k, v in kwargs.items()}
#             res = self.function(*args_, **kwargs_)
#             if res is None:
#                 return data
#             if is_ts(res):
#                 new = res
#             elif is_dict(res):
#                 new = res.get('data')
#             elif is_tuple(res):
#                 new = res[0]
#             else:
#                 raise ValueError('result of persistence data must be a timeseries or a dict \n%s '%res)
#             if new is None:
#                 both = data
#             elif is_ts(new):
#                 both = data if len(new) == 0 else pd.concat([data[data.index<new.index[0]], new])
#             else:
#                 raise ValueError('data must be None or a timeseries')                
#             if is_ts(res):
#                 return both
#             elif is_dict(res):
#                 res['data'] = both
#                 return res
#             elif is_tuple(res):
#                 return (both,) + res[1:]
#         else:
#             return self.function(*args, **kwargs)




import numpy as np
from pyg_timeseries._math import stdev_calculation, skew_calculation
from pyg_timeseries._decorators import compiled, first_, _data_state
from pyg_base import pd2np, Dict, is_num, loop_all, loop, clock

__all__ = ['ffill', 'bfill', 'fnna', 'na2v', 'v2na', 'diff', 'shift', 'ratio', 'rolling_mean', 'rolling_sum', 'rolling_rms', 'rolling_std', 'rolling_skew', 
           'diff_', 'shift_', 'ratio_', 'rolling_mean_', 'rolling_sum_', 'rolling_rms_', 'rolling_std_', 'rolling_skew_']

###############
##
## parameters
##
###############


@loop(list, dict)
def _vec(a, vec, n, value = np.nan, axis = 0):
    if len(a.shape)==2:
        shape = a.shape[1] if axis==0 else a.shape[0]
        if vec is None:
            return [np.full(abs(n), value)]*shape
        elif is_num(vec):
            return [np.array([vec])]*shape
        else:
            return vec.copy()
    else:        
        if vec is None:
            return np.full(abs(n), value)
        elif is_num(vec):
            return np.array([vec])
        return vec.copy()

@loop_all
@pd2np
@compiled
def _fnna(a, n):
    if n == 0:
        raise ValueError('n must be non-zero')
    i = 0
    _n = abs(n)
    s = (0,a.shape[0],1) if n>0 else (a.shape[0]-1,-1,-1)
    for j in range(*s):
        if not np.isnan(a[j]):
            i = i+1
            if i == _n:
                return j


@loop_all
@pd2np
@compiled
def _fnnz(a, n=1):
    if n == 0:
        raise ValueError('n must be non-zero')
    i = 0
    _n = abs(n)
    s = (0,a.shape[0],1) if n>0 else (a.shape[0]-1,-1,-1)
    for j in range(*s):
        if not np.isnan(a[j]) and a[j]!=0.:
            i = i+1
            if i == _n:
                return j
            

@compiled
def _ffill1d(a, n, prev, i):
    res = a.copy()
    for j in range(a.shape[0]):
        if np.isnan(a[j]):
            if n == 0:
                res[j] = prev
            else:
                i+=1
                if i>n:
                    res[j] = np.nan
                else:
                    res[j] = prev
        else:
            i = 0
            prev = a[j]
    return res, prev, i


def _ffill2d(a, n, prev, i):
    res = a.copy()
    if n:
        for j in range(a.shape[0]):
            mask = np.isnan(res[j])
            i[mask]+=1
            i[~mask] = 0
            prev[i>n] = np.nan
            res[j][mask] = prev[mask]
            prev = res[j].copy()
    else:
        for j in range(a.shape[0]):
            mask = np.isnan(res[j])
            res[j][mask] = prev[mask]
            prev = res[j]
    return res, prev, i

@loop(dict, list)
@pd2np
def _ffill(a, n = 0, prev = None, i = None):
    if len(a.shape) == 1:
        if i is None:
            i = 0
        if prev is None:
            prev = np.nan
        return _ffill1d(a, n, prev , i)
    else:            
        if i is None:
            i = np.zeros(a[0].shape)
        if prev is None:
            prev = i + np.nan
        return _ffill2d(a, n, prev , i)


@loop_all
@pd2np
@compiled
def _na2v(a, new = 0.0):
    res = a.copy()
    for j in range(a.shape[0]):
        if np.isnan(a[j]):
            res[j] = new
    return res

@loop_all
@pd2np
@compiled
def _v2na(a, old = 0.0, new = np.nan):
    res = a.copy()
    for j in range(a.shape[0]):
        if a[j] == old:
            res[j] = new
    return res


@loop_all
@pd2np
@compiled
def _init2v(a, n = 0, new = np.nan):
    res = a.copy()
    i = 0
    while i < a.shape[0] and n > 0:
        if ~np.isnan(a[i]):
            res[i] = new
            n = n-1
        i+=1
    return res


@loop_all
@pd2np
@compiled
def _bfill(a, limit = -1):
    """
    _bfill(np.array([np.nan, 1., np.nan])) 
    """
    res = a.copy()
    prev = np.nan
    n = limit
    for j in range(a.shape[0]-1, -1, -1):
        if np.isnan(a[j]):
            if n!=0:
                n-=1
                res[j] = prev
        else:
            n = limit
            prev = res[j]
    return res


###############
##
## bottleneck
##
###############

@loop_all
@pd2np
def _rolling_window(a, window, min_count, func, vec = None, axis = 0):
    vec = _vec(a,vec,0,axis=axis)
    mask = ~np.isnan(a)
    na = a[mask]
    n = len(na)
    if len(vec):
        na = np.concatenate([vec,na])
    w = na.shape[0] if window == 0 else window        
    res = func(na, w, min_count)
    if len(vec):
        res = res[-n:] 
    va = a.copy()
    va[mask] = res
    return va, va[-1:] if window == 0 else na[-(window-1):]


###############
##
## diff/shift/ratio
##
###############

@loop_all
@pd2np
@compiled
def _diff(a, n, vec, i, time, t = np.nan):
    """
    a = cumsum(np.array(range(10))) * 1.
    n = 1
    vec = np.array([np.nan])
    i = 0; t = np.nan
    time = np.array([0,0,1,1,1,2,2,2,2,3])

    """
    vec = vec.copy()
    _n = abs(n)
    s = (0,a.shape[0],1) if n>0 else (a.shape[0]-1,-1,-1)
    res = np.empty_like(a)
    prev = vec[i]
    for j in range(*s):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            if not (time[j] == t):
                i = (i+1) % _n
                prev = vec[i]
                # print('moving forward in time, new previous is', prev)
            t = time[j]
            res[j] = a[j] - prev
            vec[i] = a[j]
            # print('prev is ', prev, 'time is ', t, 'res is', a[j], '-', prev, '=', res[j])
    return res, vec, i, t

@loop_all
@pd2np
@compiled
def _diff1(a, vec, time, i = 0, t = np.nan):
    vec = vec.copy()
    s = (0,a.shape[0],1)
    res = np.empty_like(a)
    prev = vec[i]
    for j in range(*s):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            if not (time[j] == t):
                prev = vec[i]
            t = time[j]
            res[j] = a[j] - prev
            vec[i] = a[j]
    return res, vec, i, t


@loop_all
def _tdiff(a, n, vec, i, time = None, t = None):
    time = clock(a, time, t)
    t = 0 if t is None or np.isnan(t) else t
    return _diff(a = a, n = n, vec = vec, i = i, time = time, t = t)


@loop_all
def _tdiff1(a, vec, i = 0, time = None, t = None):
    time = clock(a, time, t)
    t = 0 if t is None or np.isnan(t) else t
    return _diff1(a = a, vec = vec, i = i, time = time, t = t)


@loop_all
@pd2np
@compiled
def _ratio(a, n, vec, i, time, t = np.nan):
    vec = vec.copy()
    _n = abs(n)
    s = (0,a.shape[0],1) if n>0 else (a.shape[0]-1,-1,-1)
    res = np.empty_like(a)
    prev = vec[i]
    for j in range(*s):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            if not (time[j] == t):
                i = (i+1) % _n
                prev = vec[i]
            t = time[j]
            res[j] = np.nan if prev == 0 else a[j] / prev 
            vec[i] = a[j]
    return res, vec, i, t

@loop_all
def _tratio(a, n, vec, i, time = None, t = None):
    time = clock(a, time, t)
    t = 0 if t is None or np.isnan(t) else t
    return _ratio(a = a, n = n, vec = vec, i = i, time = time, t = t)


@loop_all
@pd2np
@compiled
def _shift(a, n, vec, i):
    vec = vec.copy()
    res = np.empty_like(a)
    _n = abs(n)
    s = (0,a.shape[0],1) if n>0 else (a.shape[0]-1,-1,-1)
    for j in range(*s):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            res[j] = vec[i]
            vec[i] = a[j]
            i = (i+1) % _n
    return res, vec, i


@loop_all
@pd2np
@compiled
def _shift1(a, vec):
    vec = vec.copy()
    res = np.empty_like(a)
    for j in range(a.shape[0]):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            res[j] = vec[0]
            vec[0] = a[j]
    return res, vec, 0


###############
##
## rolling
##
###############


@pd2np
@compiled
def _rolling_mean(a, n, time, t0, t1, vec, i, denom, t = np.nan):
    vec = vec.copy()
    res = np.empty_like(a)
    _n = abs(n)
    s = (0,a.shape[0],1) if n>0 else (a.shape[0]-1,-1,-1)
    for j in range(*s):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            if not (time[j] == t):
                i = (i+1) % _n
                t0 +=1
            t1 += a[j]-vec[i]
            vec[i] = a[j]
            res[j] = np.nan if t0<n else t1/denom
            t = time[j]
    return res, t0, t1, vec, i, t


@loop_all
def _trolling_mean(a, n, t0, t1, vec, i, denom, t = np.nan, time = None):
    time = clock(a, time, t)
    return _rolling_mean(a, n, time = time, t = t, t0 = t0, t1 = t1, vec = vec, i = i, denom = denom)


@pd2np
@compiled
def _rolling_rms(a, n, time, t0, t2, vec, i, denom, t = np.nan):
    vec = vec.copy()
    res = np.empty_like(a)
    _n = abs(n)
    s = (0,a.shape[0],1) if n>0 else (a.shape[0]-1,-1,-1)
    for j in range(*s):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            if not (time[j] == t):
                i = (i+1) % _n
                t0 +=1
            t2 += a[j]**2-vec[i]**2
            vec[i] = a[j]
            res[j] = np.nan if t0<n else np.sqrt(t2/denom)
            t = time[j]
    return res, t0, t2, vec, i, t

@loop_all
def _trolling_rms(a, n, t0, t2, vec, i, denom, t = np.nan, time = None):
    time = clock(a, time, t)
    return _rolling_rms(a, n, time = time, t = t, t0 = t0, t2 = t2, vec = vec, i = i, denom = denom)

@pd2np
@compiled
def _rolling_std(a, n, time, t0, t1, t2, vec, i, denom, t = np.nan):
    vec = vec.copy()
    res = np.empty_like(a)
    _n = abs(n)
    s = (0,a.shape[0],1) if n>0 else (a.shape[0]-1,-1,-1)
    for j in range(*s):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            if not (time[j] == t):
                i = (i+1) % _n
                t0 +=1
            t1 += a[j]-vec[i]
            t2 += a[j]**2-vec[i]**2
            vec[i] = a[j]
            res[j] = np.nan if t0<n else stdev_calculation(t0 = n, t1 = t1, t2 = t2)
            t = time[j]
    return res, t0, t1, t2, vec, i, t

@loop_all
def _trolling_std(a, n, t0, t1, t2, vec, i, denom, t = np.nan, time = None):
    time = clock(a, time, t)
    return _rolling_std(a, n, time = time, t = t, t0 = t0, t1 = t1, t2 = t2, vec = vec, i = i, denom = denom)


@pd2np
@compiled
def _rolling_skew(a, n, time, bias, t0, t1, t2, t3, vec, i, denom, t = np.nan):
    vec = vec.copy()
    res = np.empty_like(a)
    _n = abs(n)
    s = (0,a.shape[0],1) if n>0 else (a.shape[0]-1,-1,-1)
    for j in range(*s):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            if not (time[j] == t):
                i = (i+1) % _n
                t0 +=1
            t1 += a[j]-vec[i]
            t2 += a[j]**2-vec[i]**2
            t3 += a[j]**3-vec[i]**3
            vec[i] = a[j]
            res[j] = np.nan if t0<_n else skew_calculation(t0 = _n, t1 = t1, t2 = t2, t3 = t3, bias = bias, min_sample = 1)
            t = time[j]
    return res, t0, t1, t2, t3, vec, i, t


@loop_all
def _trolling_skew(a, n, t0, t1, t2, t3, vec, i, denom, bias = False, t = np.nan, time = None):
    time = clock(a, time, t)
    return _rolling_skew(a, n = n, time = time, bias = bias, t = t, t0 = t0, t1 = t1, t2 = t2, t3 = t3, vec = vec, i = i, denom = denom)


###############
##
## API
##
###############


def fnna(a, n=1, axis = 0):
    """
    returns the index in a of the nth first non-nan.
    
    :Parameters:
    ------------
    a : array/timeseries
    n: int, optional, default = 1

    :Example:
    ---------
    >>> a = np.array([np.nan,np.nan,1,np.nan,np.nan,2,np.nan,np.nan,np.nan])
    >>> fnna(a,n=-2)
    
    """
    return _fnna(a, n, axis = axis)

def fnnz(a, n=1, axis = 0):
    """
    returns the index in a of the nth first non-nan and non-zero.
    
    :Parameters:
    ------------
    a : array/timeseries
    n: int, optional, default = 1

    :Example:
    ---------
    >>> a = np.array([[np.nan,0.,1.,np.nan,np.nan,2,np.nan,np.nan,np.nan], [0.,0.,0.,np.nan,np.nan,2,np.nan,np.nan,np.nan]]).T
    >>> #                        ^                                             ^                    ^
    >>> #                      2nd pos                                 zeros dont count      first non-zero

    >>> assert list(fnnz(a)) == [2,5] 
    
    """
    return _fnnz(a, n, axis = axis)



def bfill(a, n = -1, axis = 0):
    """
    equivalent to a.fillna('bfill'). There is no state-aware as this function is forward looking

    :Example:
    -------
    >>> from pyg import *
    >>> a = np.array([np.nan, 1., np.nan])
    >>> b = np.array([1., 1., np.nan])
    >>> assert eq(bfill(a),  b)

    :Example: pd.Series
    -------
    >>> ts = pd.Series(a, drange(-2))
    >>> assert eq(bfill(ts).values, b)
    """
    return _bfill(a, limit = n, axis = axis)


def ffill(a, n=0, axis = 0, data = None, state = None):
    """
    returns a forward filled array, up to n values forward. 
    supports state manegement which is needed if we want only nth

    
    :Parameters:
    ------------
    a : array/timeseries
        array/timeseries
    n: int, optional, default = 1
        window size
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 

    :Example:
    ---------
    >>> a = np.array([np.nan,np.nan,1,np.nan,np.nan,2,np.nan,np.nan,np.nan])
    >>> fnna(a, n=-2)
    """
    state = state or Dict(prev = None, i = None)
    return first_(_ffill(a, n=n, axis = axis, **state))

def ffill_(a, n=0, axis = 0, instate = None):
    """
    returns a forward filled array, up to n values forward. 
    supports state manegement
    
    """
    state = instate or dict(prev = None, i = None)
    return _data_state(['data', 'prev', 'i'],_ffill(a, n=n, axis = axis, **state))

ffill_.output = ['data', 'state']


def v2na(a, old = 0.0, new = np.nan):
    """
    replaces an old value with a new value (default is nan)

    :Examples:
    --------------
    >>> from pyg import *
    >>> a = np.array([1., np.nan, 1., 0.])
    >>> assert eq(v2na(a), np.array([1., np.nan, 1., np.nan]))
    >>> assert eq(v2na(a,1), np.array([np.nan, np.nan, np.nan, 0]))
    >>> assert eq(v2na(a,1,0), np.array([0., np.nan, 0., 0.]))
    
    :Parameters:
    ----------------
    a : array/timeseries
    old: float
        value to be replaced
    new : float, optional
        new value to be used, The default is np.nan.

    :Returns:
    -------
    array/timeseries

    """
    return _v2na(a, old = old, new = new)

def na2v(a, new = 0.0):
    """
    replaces a nan with a new value
    
    :Example:
    -------
    >>> from pyg import *
    >>> a = np.array([1., np.nan, 1.])
    >>> assert eq(na2v(a), np.array([1., 0.0, 1.]))
    >>> assert eq(na2v(a,1), np.array([1., 1., 1.]))
    
    :Parameters:
    ----------------
    a : array/timeseries
    new : float, optional
        DESCRIPTION. The default is 0.0.

    :Returns:
    -------
    array/timeseries

    """
    return _na2v(a, new)
    

def init2v(a, n = 0, new = np.nan):
    """
    replaces initial non-nan values with a new value. This is primarily to remove initial results of a volatile fit etc.
    
    :Example:
    -------
    >>> from pyg import *
    >>> a = np.arange(100) * 1.
    >>> assert np.sum(a[np.isnan(init2v(a,10))]) == 0+1+2+3+4+5+6+7+8+9
    
    :Parameters:
    ----------------
    a : array/timeseries
        data to be nanned
    new : float, optional
        DESCRIPTION. The default is np.nan.
    n: int
        number of entries to go to new value

    :Returns:
    -------
    array/timeseries

    """
    if n == 0:
        return a
    return _init2v(a, n, new)


def diff(a, n=1, time = None, axis = 0, data = None, state = None):
    """
    equivalent to a.diff(n) in pandas if there are no nans. If there are, we SKIP nans rather than propagate them.

    :Parameters:
    ------------
    a : array/timeseries
        array/timeseries
    n: int, optional, default = 1
        window size
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 


    :Example: : matching pandas no nan's
    ----------------------------------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> assert eq(timer(diff, 1000)(a), timer(lambda a, n=1: a.diff(n), 1000)(a))

    :Example: : nan skipping
    ----------------------------------
    >>> a = np.array([1., np.nan, 3., 9.])
    >>> assert eq(diff(a),                      np.array([np.nan, np.nan, 2.0,   6.0]))
    >>> assert eq(pd.Series(a).diff().values,   np.array([np.nan, np.nan, np.nan,6.0]))
    
    """
    if n == 0:
        return a - a
    state = state or Dict(vec = None, i = 0, t = np.nan)
    state.vec = _vec(a,state.vec, n, axis=axis)
    return first_(_tdiff1(a, time = time, axis = axis, **state) if n == 1 
                  else _tdiff(a, n = n, time = time, axis = axis, **state))

def diff_(a, n=1, time = None, axis = 0, data = None, instate = None):
    """
    returns a forward filled array, up to n values forward. 
    Equivalent to diff(a,n) but returns the full state. See diff for full details
  
    """
    if n == 0:
        return Dict(data = a - a, state = instate)
    state = instate or Dict(vec = None, i = 0, t = np.nan) 
    state.vec = _vec(a, state['vec'], n, axis=axis)
    return _data_state(['data', 'vec', 'i', 't'], 
                       _tdiff1(a, time = time, axis = axis, **state) if n == 1 
                       else _tdiff(a, n=n, time = time, axis = axis, **state))

diff_.output = ['data', 'state']
        
def shift(a, n=1, axis = 0, data = None, state = None):
    """
    Equivalent to a.shift() with support to arra
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
    n: int
        size of rolling window
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example:
    ---------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series([1.,2,3,4,5], drange(-4))
    >>> assert eq(shift(a), pd.Series([np.nan,1,2,3,4], drange(-4)))
    >>> assert eq(shift(a,2), pd.Series([np.nan,np.nan,1,2,3], drange(-4)))
    >>> assert eq(shift(a,-1), pd.Series([2,3,4,5,np.nan], drange(-4)))

    :Example: np.ndarrays
    ---------------------
    >>> assert eq(shift(a.values), shift(a).values)

    :Example: nan skipping
    ---------------------
    >>> a = pd.Series([1.,2,np.nan,3,4], drange(-4))
    >>> assert eq(shift(a), pd.Series([np.nan,1,np.nan, 2,3], drange(-4)))
    >>> assert eq(a.shift(), pd.Series([np.nan,1,2,np.nan,3], drange(-4))) # the location of the nan changes

    :Example: state management
    --------------------------
    >>> old = a.iloc[:3]
    >>> new = a.iloc[3:]
    >>> old_ts = shift_(old)
    >>> new_ts = shift(new, **old_ts)
    >>> assert eq(new_ts, shift(a).iloc[3:])
    """
    if n == 0:
        return a
    state = state or Dict(vec = None, i = 0,)
    state.vec = _vec(a, state.vec, n, axis=axis)
    return first_(_shift1(a, state.vec, axis = axis) if n == 1 else _shift(a, n, axis = axis, **state))

def shift_(a, n=1, axis = 0, instate = None):
    """
    Equivalent to shift(a,n) but returns the full state. See shift for full details
  
    """
    if n == 0:
        return Dict(data = a, state = instate)
    state = instate or Dict(vec = None, i = 0,)
    state.vec = _vec(a, state.vec, n, axis=axis)
    return _data_state(['data', 'vec', 'i'], _shift1(a, vec = state.vec, axis = axis) if n == 1 else _shift(a, n, axis = axis, **state))

shift_.output = ['data', 'state']
        
def ratio(a, n=1, time = None, data = None, state = None, axis = 0):
    """
    Equivalent to a.diff() but in log-space..
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
    n: int
        size of rolling window

    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
            
    :Example:
    ---------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series([1.,2,3,4,5], drange(-4))
    >>> assert eq(ratio(a), pd.Series([np.nan, 2, 1.5, 4/3,1.25], drange(-4)))
    >>> assert eq(ratio(a,2), pd.Series([np.nan, np.nan, 3, 2, 5/3], drange(-4)))
    """
    state = state or Dict(vec = None, i = 0, t = np.nan)
    state.vec = _vec(a, state.vec, n, axis=axis)
    return first_(_tratio(a, n, axis = axis, time = time, **state))

def ratio_(a, n=1, time = None, axis = 0, data = None, instate = None):
    state = instate or Dict(vec = None, i = 0, t = np.nan) 
    state.vec = _vec(a, state.vec, n, axis=axis)
    return _data_state(['data', 'vec', 'i', 't'], _tratio(a, n, time = time, **state))

ratio_.output = ['data', 'state']


def rolling_mean(a, n, time = None, axis = 0, data = None, state = None):
    """
    equivalent to pandas a.rolling(n).mean().
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
    n: int
        size of rolling window
    time: a sequence of rising values of time
        passage of time
    axis : int, optional
        0/1/-1. The default is 0.    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = a.rolling(10).mean(); ts = rolling_mean(a,10)
    >>> assert abs(ts-panda).max()<1e-10   

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not include the nans in the rolling calculation: it skips them.
    Since pandas rolling engine does not skip nans, they propagate. 
    In fact, having removed half the data points, rolling(10) will return 99% of nans

    >>> a[a<0.1] = np.nan
    >>> panda = a.rolling(10).mean(); ts = rolling_mean(a,10)
    >>> print('#original:', len(nona(a)), 'timeseries:', len(nona(ts)), 'panda:', len(nona(panda)), 'data points')
    >>> #original: 4534 timeseries: 4525 panda: 6 data points

    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = rolling_mean(a,10)
    >>> old_ts = rolling_mean_(old,10)
    >>> new_ts = rolling_mean(new, 10, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(rolling_mean(dict(x = a, y = a**2),10), dict(x = rolling_mean(a,10), y = rolling_mean(a**2,10)))
    >>> assert eq(rolling_mean([a,a**2],10), [rolling_mean(a,10), rolling_mean(a**2,10)])

    :Example: passage of time
    --------------------------
    >>> a = np.array([1.,2.,3.,4.])
    >>> time = np.array([0,0,1,1]) ## i.e. the first two observations are from day 0 and the next two are from day 1
    >>> rolling_mean(a,2,time)
    >>> array([nan, nan, 2.5, 3. ])  
    
    ## The first two observations are from day 0 so cannot have a mean until we get a second point
    ## The mean is then calculated from last observation in day 0 (i.e. 2) and then 3. and then with 4. since these are again, from same day
            

    """
    state = state or Dict(t0 = 0, t1 = 0., vec = None, i = 0, t = np.nan)
    state.vec = _vec(a, state.vec, n, 0., axis=axis)
    return first_(_trolling_mean(a, n, time = time, denom = n, axis = axis, **state))

def rolling_rms(a, n, time = None, axis = 0, data = None, state = None):
    """
    equivalent to pandas (a**2).rolling(n).mean()**0.5.
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
    n: int
        size of rolling window
    time: a sequence of rising values of time
        passage of time
    axis : int, optional
        0/1/-1. The default is 0.    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = (a**2).rolling(10).mean()**0.5; ts = rolling_rms(a,10)
    >>> assert abs(ts-panda).max()<1e-10   

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not include the nans in the rolling calculation: it skips them.
    Since pandas rolling engine does not skip nans, they propagate. 
    In fact, having removed half the data points, rolling(10) will return 99% of nans

    >>> a[a<0.1] = np.nan
    >>> panda = (a**2).rolling(10).mean()**0.5; ts = rolling_rms(a,10)
    >>> print('#original:', len(nona(a)), 'timeseries:', len(nona(ts)), 'panda:', len(nona(panda)), 'data points')
    >>> #original: 4534 timeseries: 4525 panda: 6 data points

    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = rolling_rms(a,10)
    >>> old_ts = rolling_rms_(old,10)
    >>> new_ts = rolling_rms(new, 10, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(rolling_rms(dict(x = a, y = a**2),10), dict(x = rolling_rms(a,10), y = rolling_rms(a**2,10)))
    >>> assert eq(rolling_rms([a,a**2],10), [rolling_rms(a,10), rolling_rms(a**2,10)])

    :Example: passage of time
    --------------------------
    >>> a = np.array([1.,2.,3.,4.])
    >>> time = np.array([0,0,1,1]) ## i.e. the first two observations are from day 0 and the next two are from day 1
    >>> (rolling_rms(a,2,time) ** 2) * 2
    >>> array([nan, nan, 13., 20.])   == array([nan, nan, 4 + 9, 4 + 16])
    
    ## The first two observations are from day 0 so cannot have a mean until we get a second point
    ## The rms is then calculated from last observation in day 0 (i.e. 2) and then 3. and then with 4. since these are again, from same day
            

    """
    state = state or Dict(t0 = 0, t2 = 0., vec = None, i = 0, t = np.nan)
    state.vec = _vec(a, state.vec, n, 0., axis=axis)
    return first_(_trolling_rms(a, n, time = time, denom = n, axis = axis, **state))

def rolling_sum(a, n, time = None, axis = 0, data = None, state = None):
    """
    equivalent to pandas a.rolling(n).sum().
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
    n: int
        size of rolling window
    time: a sequence of rising values of time
        passage of time
    axis : int, optional
        0/1/-1. The default is 0.    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = a.rolling(10).sum(); ts = rolling_sum(a,10)
    >>> assert abs(ts-panda).max()<1e-10   

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not include the nans in the rolling calculation: it skips them.
    Since pandas rolling engine does not skip nans, they propagate. 
    In fact, having removed half the data points, rolling(10) will return 99.9% nans

    >>> a[a<0.1] = np.nan
    >>> panda = a.rolling(10).sum(); ts = rolling_sum(a,10)
    >>> print('#original:', len(nona(a)), 'timeseries:', len(nona(ts)), 'panda:', len(nona(panda)), 'data points')
    >>> #original: 4534 timeseries: 4525 panda: 2 data points

    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = rolling_sum(a,10)
    >>> old_ts = rolling_sum_(old,10)
    >>> new_ts = rolling_sum(new, 10, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(rolling_sum(dict(x = a, y = a**2),10), dict(x = rolling_sum(a,10), y = rolling_sum(a**2,10)))
    >>> assert eq(rolling_sum([a,a**2],10), [rolling_sum(a,10), rolling_sum(a**2,10)])

    :Example: passage of time
    --------------------------
    >>> a = np.array([1.,2.,3.,4.])
    >>> time = np.array([0,0,1,1]) ## i.e. the first two observations are from day 0 and the next two are from day 1
    >>> rolling_sum(a,2,time) 
    >>> array([nan, nan, 5., 6.])   == array([nan, nan, 2+3, 2+4])
    
    ## The first two observations are from day 0 so cannot have a mean until we get a second point
    ## The sum is then calculated from last observation in day 0 (i.e. 2) and then 3. and then with 4. since these are again, from same day
    """
    state = state or Dict(t0 = 0, t1 = 0., vec = None, i = 0, t = np.nan)
    state.vec = _vec(a, state.vec, n, 0., axis=axis)
    return first_(_trolling_mean(a, n, time = time, denom = 1, axis = axis, **state))

def rolling_std(a, n, time = None, axis = 0, data = None, state = None):
    """
    equivalent to pandas a.rolling(n).std().
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
    n: int
        size of rolling window
    time: a sequence of rising values of time
        passage of time
    axis : int, optional
        0/1/-1. The default is 0.    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = a.rolling(10).std(); ts = rolling_std(a,10)
    >>> assert abs(ts-panda).max()<1e-10   

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not include the nans in the rolling calculation: it skips them.
    Since pandas rolling engine does not skip nans, they propagate. 
    In fact, having removed half the data points, rolling(10) will return 99.9% nans

    >>> a[a<0.1] = np.nan
    >>> panda = a.rolling(10).std(); ts = rolling_std(a,10)
    >>> print('#original:', len(nona(a)), 'timeseries:', len(nona(ts)), 'panda:', len(nona(panda)), 'data points')
    >>> #original: 4534 timeseries: 4525 panda: 2 data points

    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = rolling_std(a,10)
    >>> old_ts = rolling_std_(old,10)
    >>> new_ts = rolling_std(new, 10, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(rolling_std(dict(x = a, y = a**2),10), dict(x = rolling_std(a,10), y = rolling_std(a**2,10)))
    >>> assert eq(rolling_std([a,a**2],10), [rolling_std(a,10), rolling_std(a**2,10)])

    :Example: passage of time
    --------------------------
    >>> a = np.array([1.,2.,3.,4.])
    >>> time = np.array([0,0,1,1]) ## i.e. the first two observations are from day 0 and the next two are from day 1
    >>> 2*(rolling_std(a,2,time) ** 2)
    >>> array([nan, nan, 1., 4.])
    
    ## The first two observations are from day 0 so cannot have a mean until we get a second point
    ## The sum is then calculated from last observation in day 0 (i.e. 2) and then 3. and then with 4. since these are again, from same day
    """    
    state = state or Dict(t0 = 0, t1 = 0, t2 = 0., vec = None, i = 0, t = np.nan)
    state.vec = _vec(a, state.vec, n, 0., axis=axis)
    return first_(_trolling_std(a, n, time = time, denom = n, axis = axis, **state))

def rolling_skew(a, n, bias = False, time = None, axis = 0, data = None, state = None):
    """
    equivalent to pandas a.rolling(n).skew().
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
    n: int
        size of rolling window
    time: a sequence of rising values of time
        passage of time
    bias: 
        affects the skew calculation definition, see scipy documentation for details.
    axis : int, optional
        0/1/-1. The default is 0.    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = a.rolling(10).skew(); ts = rolling_skew(a,10)
    >>> assert abs(ts-panda).max()<1e-10   

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not include the nans in the rolling calculation: it skips them.
    Since pandas rolling engine does not skip nans, they propagate. 
    In fact, having removed half the data points, rolling(10) will return 99.9% nans

    >>> a[a<0.1] = np.nan
    >>> panda = a.rolling(10).skew(); ts = rolling_skew(a,10)
    >>> print('#original:', len(nona(a)), 'timeseries:', len(nona(ts)), 'panda:', len(nona(panda)), 'data points')
    >>> #original: 4534 timeseries: 4525 panda: 2 data points

    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = rolling_skew(a,10)
    >>> old_ts = rolling_skew_(old,10)
    >>> new_ts = rolling_skew(new, 10, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(rolling_skew(dict(x = a, y = a**2),10), dict(x = rolling_skew(a,10), y = rolling_skew(a**2,10)))
    >>> assert eq(rolling_skew([a,a**2],10), [rolling_skew(a,10), rolling_skew(a**2,10)])
    """
    state = state or Dict(t0 = 0, t1 = 0, t2 = 0., t3 = 0, vec = None, i = 0, t = np.nan)
    state.vec = _vec(a, state.vec, n, 0., axis=axis)
    return first_(_trolling_skew(a, n, time = time, bias = bias, denom = n, axis = axis, **state))


def rolling_mean_(a, n, time = None, axis = 0, data = None, instate = None):
    """
    Equivalent to rolling_mean(a) but returns also the state variables t0,t1 etc. 
    For full documentation, look at rolling_mean.__doc__
    """
    state = instate or Dict(t0 = 0, t1 = 0., vec = None, i = 0, t = np.nan)
    state.vec = _vec(a, state.vec, n, 0., axis=axis)
    return _data_state(['data','t0','t1', 'vec','i', 't'],_trolling_mean(a, n, time = time, denom = n, axis = axis, **state))

rolling_mean_.output = ['data','state']

def rolling_rms_(a, n, time = None, axis = 0, data = None, instate = None):
    """
    Equivalent to rolling_rms(a) but returns also the state variables t0,t1 etc. 
    For full documentation, look at rolling_rms.__doc__
    """
    state = instate or Dict(t0 = 0, t2 = 0., vec = None, i = 0, t = np.nan)
    state.vec = _vec(a, state.vec, n, 0., axis=axis)
    return _data_state(['data','t0','t2', 'vec','i', 't'],_trolling_rms(a, n, time = time, denom = n, axis = axis, **state))

rolling_rms_.output = ['data','state']

def rolling_sum_(a, n, time = None, axis = 0, data = None, instate = None):
    """
    Equivalent to rolling_sum(a) but returns also the state variables t0,t1 etc. 
    For full documentation, look at rolling_sum.__doc__
    """
    state = instate or Dict(t0 = 0, t1 = 0., vec = None, i = 0, t = np.nan)
    state.vec = _vec(a, state.vec, n, 0., axis=axis)
    return _data_state(['data','t0','t1', 'vec','i', 't'], _trolling_mean(a, n, denom = 1, axis = axis, **state))

rolling_sum_.output = ['data','state']

def rolling_std_(a, n, time = None, axis = 0, data = None, instate = None):
    """
    Equivalent to rolling_std(a) but returns also the state variables t0,t1 etc. 
    For full documentation, look at rolling_std.__doc__
    """
    state = instate or Dict(t0 = 0, t1 = 0, t2 = 0., vec = None, i = 0, t = np.nan)
    state.vec = _vec(a, state.vec, n, 0., axis=axis)
    return _data_state(['data','t0', 't1', 't2', 'vec', 'i', 't'],_trolling_std(a, n, time = time, denom = n, axis = axis, **state))


rolling_std_.output = ['data','state']

def rolling_skew_(a, n, time = None, bias = False, axis = 0, data = None, instate = None):
    """
    Equivalent to rolling_skew(a) but returns also the state variables t0,t1 etc. 
    For full documentation, look at rolling_skew.__doc__
    """
    state = instate or Dict(t0 = 0, t1 = 0, t2 = 0., t3 = 0, vec = None, i = 0, t = np.nan)
    state.vec = _vec(a, state.vec, n, 0., axis=axis)
    return _data_state(['data','t0', 't1', 't2', 't3', 'vec','i', 't'], _trolling_skew(a, n, time = time, bias = bias, denom = n, axis = axis, **state))

rolling_skew_.output = ['data','state']



from pyg_base import Dict, is_df, is_nums, is_pd, dictable
import numpy as np
import pandas as pd
from pyg_timeseries._index import df_concat
from pyg_timeseries._rolling import init2v, ffill
from pyg_timeseries._ewm import ewma
        


def shape1(value):
    """
    >>> assert shape1(3) == 0
    >>> assert shape1(np.arange(10)) == 0
    >>> assert shape1(np.random.normal(0,1,(10,20))) == 20
    """
    shape = getattr(value, 'shape', ())
    return shape[1] if len(shape)>1 else 0

def _rms_combine(signals, weights, days = 1024, max_leverage = 5):
    """
    
    :Theory:
    -------
    Suppose we have a collctions of RMS1 random variables sig_i with relative weights w_i and we want to aggregate:
        
    >>>  X = \sum w_i sig_i      # Note that w_i can be negative
    
    We would like to normalize X so that X is also RMS1. We use a single factor model:
    
    >>> E(sig_i * sig_j) = rho   # instantenously for all i <> j
    >>> E(X^2) = \sum w_i^2 + rho \sum_{i<>j} w_i w_j 
    
    So 
    
    >>> rho^ = (X^2 - \sum w_i^2) / \sum_{i<>j} w_i w_j  

    is a point estimator for rho at time t and we use recent history as the full history estimator:
    
    >>> rho = ewma(rho^, days)   
    
    :Risk and max leverage:
    -----------------------
    Given rho, we can now estimate X variance based on current sig and weights
    
    >>> var(X) = E(X^2) = \sum w_i^2 + rho \sum_{i<>j} w_i w_j 
    >>> vol(X) = var(X)^0.5    
    
    Now we want to estimate leverage: This is how much absolute risk we have versus risk of X
    
    >>> sum of risks = \sum |w_i|
    >>> risk of sums = vol(X)
    >>> leverage = \sum|w_i| / vol(X)
    
    At times, we want to ensure our vol estimation does not drop to low. So we cap cap maximum leverage. 
    We then use the capped leverage to re-estimate the floored vol. 
        
    >>> vol = \sum|w_i| / leverage 
    >>> mult = 1/vol
    
    Finally, we provide mult, the multiplier that satisfies mult X ~ RMS1:
    
    :Example:
    ---------
    from pyg import * 
    
    >>> dates = drange(-9999)
    >>> signals = ewmxo(np.random.normal(0,1,(10000,10)), 18, 54, vol = 30)
    >>> weights = np.array([np.arange(1,11)]*10000) * 1. / 45.
    >>> pd.DataFrame(signals, dates).plot()
    >>> days = 1024
    >>> max_leverage = 5

    Parameters
    ----------
    signals : TYPE
        DESCRIPTION.
    weights : TYPE
        DESCRIPTION.
    cor_days : TYPE, optional
        DESCRIPTION. The default is 1024.
    max_leverage : TYPE, optional
        DESCRIPTION. The default is 5.

    Returns
    -------
    None.

    """
    mask = np.isnan(signals)
    all_nan = np.min(mask, axis=1)
    signals[mask] = 0.0
    weights[mask] = 0.0
    ws = weights * signals
    ws2 = ws ** 2
    w2 = weights ** 2    
    ws_sum = np.sum(ws, axis=1)
    ws2_sum = np.sum(ws2, axis=1)
    w2_sum = np.sum(w2, axis=1)
    w_abs = np.abs(weights)
    w_abs_sum = np.sum(w_abs, axis=1)
    w_ij = w_abs_sum ** 2 - w2_sum # this has no signal in it
    w_ij[w_ij ==0] = np.nan  
    rho_point = (ws2_sum - w2_sum) / w_ij
    rho_point[all_nan] = np.nan
    rho = ewma(init2v(rho_point, 0.3 * days, 0.0), days)
    rho = np.minimum(rho, 1)
    rho = np.maximum(rho, -1)
    var = np.maximum(w2_sum + rho * w_ij, 0)
    vol = np.sqrt(var)
    leverage = w_abs_sum / vol
    if max_leverage>0:
        leverage = np.minimum(leverage, max_leverage)
        vol = w_abs_sum / leverage
    mult = 1/vol
    signal = ws_sum / vol
    return dict(data = signal, rho = rho, mult = mult, vol = vol, var = var, leverage = leverage)
    



def rms_combine(signals, weights = None, days = 1024, max_leverage = 5, join = 'outer', method = 'ffill'):
    """
    >>> dt1 = drange(2000,2020, '1b')
    >>> sig1 = pd.DataFrame(ewmxo(np.random.normal(0,1,(len(dt1),10)), 18, 54, vol = 30) , dt1)
    >>> sig2 = pd.DataFrame(ewmxo(np.random.normal(0,1,(len(dt1),10)), 18, 54, vol = 30) , dt1)
    >>> signals = [sig1, sig2]    
    >>> rms_combine(signals)
    >>> rms_combine(sig1)

    Parameters
    ----------
    signals : dataframe/list of dataframes
        list of RMS signals
    weights : 
        list of weights
    days : int
        Days for calculating correlation. The default is 1024.
    join : str, optional
        how join is done. The default is 'outer'.
        
        
    :Example: single data frame
    ---------------------------
    >>> from pyg import * 
    >>> dates = drange(2000,2020, '1b')
    >>> signals = pd.DataFrame(ewmxo(np.random.normal(0,1,(len(dates),10)), 18, 54, vol = 30) , dates)
    >>> signals[np.random.normal(0,1,(len(dates),10))>2] = np.nan
    >>> method = 'ffill'; days = 1024
    >>> join = 'outer'
    >>> res = rms_combine(signals)
    >>> res.data.plot()
    
    :Example: single np.ndarray
    ---------
    >>> signals = ewmxo(np.random.normal(0,1,(len(dates),10)), 18, 54, vol = 30)
    >>> res2 = rms_combine(signals)
    
    
    :Example: list of signals
    -------------------------
    >>> rtn = pd.Series(np.random.normal(0,1,len(dates)), dates)
    >>> sig1 = pd.Series(ewmxo(rtn, 5, 15, vol = 30) , dates)
    >>> sig2 = pd.Series(ewmxo(rtn, 10, 30, vol = 30) , dates)
    >>> sig3 = pd.Series(ewmxo(rtn, 20, 60, vol = 30) , dates)
    >>> signals = [sig1, sig2, sig3]    
    >>> weights = None    
    >>> res3 = rms_combine(signals)
    >>> res3.rho.plot()
    
    Returns
    -------
    None.

    """
    if is_df(signals): # we have a single dataframe of multiple signals
        if method == 'ffill':
            signals = ffill(signals)

        index = signals.index
        columns = signals.columns
        s = signals.values.astype(float)

        if weights is None:
            w = 0 * s + 1
        elif isinstance(weights, list):
            if is_nums(weights):
                w = np.array([weights]*len(signals))
            else:
                w = pd.concat(weights, axis = 1).reindex(index, method = 'ffill').values
        elif is_pd(weights):
            w = weights[columns].reindex(index, method = 'ffill').values
        else:
            raise ValueError('dont know how to interpret weights')
        w = w.astype(float)
        res = _rms_combine(s, w, days = days, max_leverage = max_leverage)
        res = {key: pd.Series(value, index) for key, value in res.items()}
        return Dict(res)

    if isinstance(signals, np.ndarray):
        if method == 'ffill':
            signals = ffill(signals)
        s = signals.astype(float)
        
        if weights is None:
            w = 0 * s + 1
        elif isinstance(weights, list):
            if is_nums(weights):
                w = np.array([weights]*len(signals))
            else:
                w = np.array([np.arange(10), np.arange(10)]).T
        elif is_pd(weights):
            w = weights.ffill().values
        else:
            raise ValueError('dont know how to interpret weights')            
        w = w.astype(float)
        return _rms_combine(s, w, days = days, max_leverage = max_leverage)
        

    if isinstance(signals, list):
        ns = set([sig.shape[1] for sig in signals if len(getattr(sig, 'shape', ())) == 2]) - {1,}
        if len(ns) > 1:
            raise ValueError('multiple dataframes with different widths %s'%ns)
        elif len(ns) == 0: # single calculations, this is simple
            sigs = pd.concat(signals, axis=1, join = join)
            return rms_combine(sigs, weights = weights, days = days, max_leverage = max_leverage, join = join, method = method)
        elif len(ns) == 1: # we are combining listed signals, each operates on a multi-column asset
            n = list(ns)[0]
            if weights is None:
                weights = [1.] * len(signals)
            cols = set([tuple(sig.columns) for sig in signals if shape1(sig) == n] + [tuple(wgt.columns) for wgt in weights if shape1(wgt) == n])
            if len(cols)>1:
                raise ValueError('all columns within the list must be the same %s'%cols)
            columns = list(list(cols)[0])
            res = []
            for i in range(n):
                sigs = [sig[i] if shape1(sig) == n else sig for sig in signals]
                wgts = [wgt[i] if shape1(wgt) == n else wgt for wgt in weights]
                res.append(rms_combine(sigs, wgts, days = days, max_leverage = max_leverage, join = join, method = method))            
            res = Dict(dictable(res)).do(lambda value: df_concat(value, columns))
            return res
        
    if isinstance(signals, dict):
        wgts = [weights[key] for key in signals if key in signals] if isinstance(weights, dict) else weights
        sigs = list(signals.values())
        return rms_combine(sigs, wgts, days = days, max_leverage = max_leverage, join = join, method = method)
    
    raise ValueError('dont know how to interpret signals')
    


import numpy as np
import pandas as pd
from pyg_timeseries._math import stdev_calculation, skew_calculation, cor_calculation
from pyg_timeseries._decorators import compiled, first_, _data_state
from pyg_timeseries._rolling import _vec
from pyg_base import pd2np, loop_all, is_num, is_nums, as_list, is_ts, is_pd, logger


__all__ = ['ts_std', 'ts_mean', 'ts_skew', 'ts_count', 'ts_min', 'ts_max', 'ts_rms', 'ts_median',  'ts_sum',
           'ts_std_', 'ts_mean_', 'ts_skew_', 'ts_count_', 'ts_min_', 'ts_max_', 'ts_rms_', 'ts_sum_']

###############
##
## calculators
##
###############

@loop_all
@pd2np
def _median(a):
    return np.median(a[~np.isnan(a)])

def ts_median(a, axis = 0):
    return _median(a, axis = axis)

@pd2np
@compiled
def _moments(a, vec):
    res = vec.copy()
    for i in range(a.shape[0]):
        if not np.isnan(a[i]):
            v = 1
            for j in range(vec.shape[0]-1):
                res[j] += v
                v = v  * a[i]
            res[-1] += v
    return res


@pd2np
@compiled
def _min(a, m = np.inf):
    for i in range(a.shape[0]):
        if not np.isnan(a[i]) and a[i]<m:
            m = a[i]
    return m

@pd2np
@compiled
def _max(a, m = np.inf):
    for i in range(a.shape[0]):
        if not np.isnan(a[i]) and a[i]>m:
            m = a[i]
    return m

@loop_all
def _ts_min(a, m = None):
    if m is None or np.isnan(m):
        m = np.inf
    m = _min(a, m)
    if np.isinf(m):
        return np.nan, m
    else:
        return m, m

def ts_min(a, axis = 0, data = None, state = None):
    """
    ts_min(a) is equivalent to pandas a.min()
    """
    state = state or {}
    return first_(_ts_min(a, axis = axis, **state))
    
def ts_min_(a, axis = 0, data = None, instate = None):
    """
    ts_min(a) is equivalent to pandas a.min()
    """
    state = instate or {}
    res = _ts_min(a, axis = axis, **state)
    return _data_state(['data', 'm'], res)

ts_min_.output = ['data', 'state']

@loop_all
def _ts_max(a, m = None):
    if m is None or np.isnan(m):
        m = -np.inf
    m = _max(a, m)
    if np.isinf(m):
        return np.nan, m
    else:
        return m, m
    
def ts_max(a, axis = 0, data = None, state = None):
    """
    ts_max(a) is equivalent to pandas a.min()
    """
    state = state or {}
    return first_(_ts_max(a, axis = axis, **state))

def ts_max_(a, axis = 0, data = None, instate = None):
    """
    ts_max(a) is equivalent to pandas a.min()
    """
    state = instate or {}
    res = _ts_max(a, axis = axis, **state)
    return _data_state(['data', 'm'], res)

ts_max_.output = ['data', 'state']

def _zip(value):
    return _data_state(['data', 'vec'], value)

@loop_all
def _ts_count(a, vec = None):
    vec = _vec(vec,1,0)
    vec = _moments(a, vec)
    return vec[0], vec
    

def ts_count(a, axis = 0, data = None, state = None):
    """
    ts_count(a) is equivalent to a.count() (though slightly slower)
    
    - supports numpy arrays 
    - skips nan
    - supports state management
    
    :Example: pandas matching
    -----------------------------------
    >>> # create sample data:
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999)); a[a>0] = np.nan
    >>> assert ts_count(a) == a.count()

    :Example: numpy 
    -----------------------------------
    >>> assert ts_count(a.values) == ts_count(a)

    :Example: state management
    -------------------------------------------
    >>> old = ts_count_(a.iloc[:2000])
    >>> new = ts_count(a.iloc[2000:], state = old.state)
    >>> assert new == ts_count(a)

    """
    state = state or {}
    return first_(_ts_count(a, axis = axis, **state))

def ts_count_(a, axis = 0, data = None, instate = None):
    """
    ts_count_(a) is equivalent to ts_count(a) except vec is also returned.
    See ts_count for full documentation        
    """
    state = instate or {}
    return _zip(_ts_count(a, axis = axis, **state))

@loop_all
def _ts_sum(a, vec = None):
    vec = _vec(vec,2,0.)
    vec = _moments(a, vec)
    return vec[1], vec
    
def ts_sum(a, axis = 0, data = None, state = None):
    """
    ts_sum(a) is equivalent to a.sum()
    
    - supports numpy arrays 
    - handles nan
    - supports state management

    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.

    data: None
        unused at the moment. Allow code such as func(live, **func_(history)) to work

    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
    

    :Example: pandas matching
    -----------------------------------
    >>> # create sample data:
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999)); a[a>0] = np.nan
    >>> assert ts_sum(a) == a.sum()

    :Example: numpy 
    -----------------------------------
    >>> assert ts_sum(a.values) == ts_sum(a)

    :Example: state management
    -------------------------------------------
    >>> old = ts_sum_(a.iloc[:2000])
    >>> new = ts_sum(a.iloc[2000:], vec = old.vec)
    >>> assert new == ts_sum(a)
    """
    state = state or {}
    if is_num(a) or (isinstance(a, list) and is_nums(a)):
        a = np.array(as_list(a))
    return first_(_ts_sum(a, axis = axis, **state))

def ts_sum_(a, axis = 0, data = None, instate = None):
    """
    ts_sum_(a) is equivalent to ts_sum(a) except vec is also returned.
    See ts_sum for full documentation        
    """
    state = instate or {}
    if is_num(a) or (isinstance(a, list) and is_nums(a)):
        a = np.array(as_list(a))
    return _zip(_ts_sum(a, axis = axis, **state))

@loop_all
def _ts_mean(a, vec = None):
    vec = _vec(vec,2,0.)
    vec = _moments(a, vec)
    return vec[1]/vec[0], vec

def ts_mean(a, axis = 0, data = None, state = None):
    """
    ts_mean(a) is equivalent to a.mean()
    
    - supports numpy arrays 
    - handles nan
    - supports state management
    - pandas is actually faster on count
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.

    data: None
        unused at the moment. Allow code such as func(live, **func_(history)) to work

    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 

    :Example: pandas matching
    -----------------------------------
    >>> # create sample data:
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999)); a[a>0] = np.nan
    >>> assert ts_mean(a) == a.mean()

    :Example: numpy 
    -----------------------------------
    >>> assert ts_mean(a.values) == ts_mean(a)

    :Example: state management
    -------------------------------------------
    >>> old = ts_mean_(a.iloc[:2000])
    >>> new = ts_mean(a.iloc[2000:], vec = old.vec)
    >>> assert new == ts_mean(a)
    """
    state = state or {}
    if is_num(a) or (isinstance(a, list) and is_nums(a)):
        a = np.array(as_list(a))
    return first_(_ts_mean(a, axis = axis, **state))

def ts_mean_(a, axis = 0, data = None, instate = None):
    """
    ts_mean_(a) is equivalent to ts_mean(a) except vec is also returned.
    See ts_mean for full documentation        
    """
    state = instate or {}
    if is_num(a) or (isinstance(a, list) and is_nums(a)):

        a = np.array(as_list(a))
    return _zip(_ts_mean(a, axis = axis, **state))


@loop_all
@pd2np
@compiled
def _ts_cor(a, b, min_sample, vec):
    res = vec.copy()
    for i in range(a.shape[0]):
        if not np.isnan(a[i]) and not np.isnan(b[i]):
            res[0] += 1
            res[1] += a[i]
            res[2] += a[i]**2
            res[3] += b[i]
            res[4] += b[i]**2
            res[5] += a[i]*b[i]
    return cor_calculation(t0 = res[0], a1 = res[1], a2 = res[2], b1 = res[3], b2 = res[4], ab = res[5], min_sample = min_sample), res
    

def ts_cor(a, b, min_sample = 3, axis = 0, data = None, state = None):
    """
    ts_cor(a) is equivalent to a.cor()[0][1]
    
    - supports numpy arrays 
    - handles nan
    - supports state management
    
    :Example: matching pandas
    -------------------------
    >>> # create sample data:
    >>> from pyg_timeseries import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999)); a[a>0.5] = np.nan
    >>> b = pd.Series(np.random.normal(0,1,10000), drange(-9999)); b[b>0.5] = np.nan
    >>> state = data = None; min_sample = 3; axis = 0
    >>> df = pd.concat([a,b], axis=1)
    >>> assert abs(df.corr()[0][1] - ts_cor(a, b))<1e-10

    :Example: slightly faster than pandas
    -------------------------------------
    %timeit ts_cor(a, b)
    245 µs ± 6.43 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)
    %timeit df.corr()[0][1]
    575 µs ± 13 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)    

    :Example: numpy 
    -----------------------------------
    >>> assert ts_cor(a.values, b.values) == ts_cor(a,b)

    :Example: state management
    -------------------------------------------
    >>> old = ts_std_(a.iloc[:2000])
    >>> new = ts_std(a.iloc[2000:], vec = old.vec)
    >>> assert new == ts_std(a)

    """
    state = state or dict(vec = _vec(None,6,0.))
    rtn = first_(_ts_cor(a, b, min_sample = min_sample, **state))
    return rtn

def ts_cor_(a, b, min_sample = 3, axis = 0, data = None, instate = None):
    """
    ts_cor_(a, b) is equivalent to ts_cor(a,b) except vec is also returned.
    See ts_std for full documentation    
    """
    state = instate or dict(vec = _vec(None,6,0.))
    if is_num(a) or (isinstance(a, list) and is_nums(a)):
        a = np.array(as_list(a))
    if is_num(b) or (isinstance(b, list) and is_nums(b)):
        b = np.ndarray(as_list(b))
    return _zip(_ts_cor(a, b, min_sample, axis = axis, **state))


@loop_all
def _ts_std(a, vec = None):
    vec = _vec(vec,3,0.)
    vec = _moments(a, vec)
    return stdev_calculation(*vec), vec


def ts_std(a, axis = 0, data = None, state = None):
    """
    ts_std(a) is equivalent to a.std()
    
    - supports numpy arrays 
    - handles nan
    - supports state management
    
    >>> # create sample data:
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999)); a[a>0] = np.nan

    :Example: pandas matching
    -----------------------------------
    >>> assert abs(ts_std(a) - a.std())<1e-13

    :Example: numpy 
    -----------------------------------
    >>> assert ts_std(a.values) == ts_std(a)

    :Example: state management
    -------------------------------------------
    >>> old = ts_std_(a.iloc[:2000])
    >>> new = ts_std(a.iloc[2000:], vec = old.vec)
    >>> assert new == ts_std(a)

    """
    state = state or {}
    if is_num(a) or (isinstance(a, list) and is_nums(a)):
        a = np.array(as_list(a))
    return first_(_ts_std(a, axis = axis, **state))

def ts_std_(a, axis = 0, data = None, instate = None):
    """
    ts_std_(a) is equivalent to ts_std(a) except vec is also returned.
    See ts_std for full documentation    
    """
    state = instate or {}
    if is_num(a) or (isinstance(a, list) and is_nums(a)):
        a = np.array(as_list(a))
    return _zip(_ts_std(a, axis = axis, **state))

@loop_all
def _ts_rms(a, vec = None):
    vec = _vec(vec,3,0.)
    vec = _moments(a, vec)
    return np.sqrt(vec[2]/vec[0]), vec
    


def ts_rms(a, axis = 0, data = None, state = None):
    """
    ts_rms(a) is equivalent to (a**2).mean()**0.5
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.

    data: None
        unused at the moment. Allow code such as func(live, **func_(history)) to work

    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 

    - supports numpy arrays 
    - handles nan
    - supports state management
    
    >>> # create sample data:
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999)); a[a>0] = np.nan

    :Example: pandas matching
    -----------------------------------
    >>> assert abs(ts_std(a) - a.std())<1e-13

    :Example: numpy 
    -----------------------------------
    >>> assert ts_std(a.values) == ts_std(a)

    :Example: state management
    -------------------------------------------
    >>> old = ts_rms_(a.iloc[:2000])
    >>> new = ts_rms(a.iloc[2000:], vec = old.vec)
    >>> assert new == ts_rms(a)

    """
    state = state or {}
    if is_num(a) or (isinstance(a, list) and is_nums(a)):
        a = np.array(as_list(a))
    return first_(_ts_rms(a, axis = axis, **state))

def ts_rms_(a, axis = 0, data = None, instate = None):
    """
    ts_rms_(a) is equivalent to ts_rms(a) except it also returns vec
    see ts_rms for full documentation
    """
    state = instate or {}
    if is_num(a) or (isinstance(a, list) and is_nums(a)):
        a = np.array(as_list(a))
    return _zip(_ts_rms(a, axis = axis, **state))


@loop_all
def _ts_skew(a, bias = False, min_sample = 0.25, vec = None):
    vec = _vec(vec,4,0.)
    vec = _moments(a, vec)
    return skew_calculation(t0 = vec[0], t1 = vec[1], t2 = vec[2], t3 = vec[3], bias = bias, min_sample = min_sample), vec

def ts_skew(a, bias = False, min_sample = 0.25, axis = 0, data = None, state = None):
    """
    ts_skew(a,0) is equivalent to a.skew()
    
    - supports numpy arrays 
    - handles nan
    - faster than pandas
    - supports state management
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.
        
    min_sample: float, optional
        This refers to the denominator when we calculate the skew. Over time, the deonimator converges to 1 but initially, it is small. 
        Also, if there is a gap in the data, older datapoints weight may have decayed while there are not enough "new point". 
        min_sample ensures that in both cases, if denominator<0.25 )(default value) we return nan.

    data: None
        unused at the moment. Allow code such as func(live, **func_(history)) to work

    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 

    :Example: pandas matching
    -----------------------------------
    >>> # create sample data:
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999)); a[a>0] = np.nan
    >>> assert abs(ts_skew(a, 0) - a.skew())<1e-13

    :Example: numpy 
    -----------------------------------
    >>> assert ts_skew(a.values) == ts_skew(a)

    :Example: state management
    -------------------------------------------
    >>> old = ts_skew_(a.iloc[:2000])
    >>> new = ts_skew(a.iloc[2000:], vec = old.vec)
    >>> assert new == ts_skew(a)
    """
    state = state or {}
    if is_num(a) or (isinstance(a, list) and is_nums(a)):
        a = np.array(as_list(a))
    return first_(_ts_skew(a, bias = bias, min_sample = min_sample, axis = axis, **state))
    
def ts_skew_(a, bias = False, min_sample = 0.25, axis = 0, data = None, instate = None):
    """
    ts_skew_(a) is equivalent to ts_skew except vec is also returned.
    See ts_skew for full details
    """
    state = instate or {}
    if is_num(a) or (isinstance(a, list) and is_nums(a)):
        a = np.array(as_list(a))
    return _zip(_ts_skew(a, bias = bias, min_sample = min_sample, axis = axis, **state))



ts_cor_.output = ['data', 'state']
ts_min_.output = ['data', 'state']
ts_max_.output = ['data', 'state']
ts_count_.output = ['data', 'state']
ts_sum_.output = ['data', 'state']
ts_mean_.output = ['data', 'state']
ts_rms_.output = ['data', 'state']
ts_std_.output = ['data', 'state']
ts_skew_.output = ['data', 'state']


def ts_interval(ts, min_freq = 0.5):
    """
    returns the most common time difference in the timeseries

    :Example:
    ----------
    >>> ts = pd.Series(range(100), drange(-200,0,'1b')[:100])
    >>> assert ts_interval(ts) == datetime.timedelta(1)

    >>> ts = pd.Series(range(100), drange(-1,0,'5n')[:100])
    >>> assert ts_interval(ts) ==  datetime.timedelta(minutes = 5)
    
    >>> ts = pd.Series(range(10000), range(10000))
    >>> ts = ts[ts % 2 != 0]
    >>> ts = ts[ts % 3 != 0]
    >>> ts = ts[ts % 5 != 0]
    >>> assert ts_interval(ts) == 4
    
    
    """
    if len(ts)<2:
        return None
    v = ts.index if is_pd(ts) else ts
    intervals = pd.Series(v, v).diff().iloc[1:]
    res = intervals.mode()[0]
    freq = len(intervals[intervals.values==res]) / len(intervals)
    if freq>=min_freq:
        return res
    else: 
        logger.warning('timeseries is irregular, with mode frequency less than %1.2f. returning median interval instead'%freq)
        return intervals.median()
    


from pyg_base import Dict, is_num, sub_, div_
from pyg_timeseries._ewm import ewma_, ewmstd_, ewmrms_
from pyg_timeseries._rolling import v2na, diff_
from pyg_timeseries._expanding import cumsum_

def _frac(days):
    return 1/(1+days) if days>1 else days
        
def ou_factor(fast, slow):
    """
    OU factor for momentum predictions.
    Calculatates the variance of an OU process defined as ewma(dB, fast) - ewma(dB, slow) if dB is a standard Brownian Motion
    
    Suppose 

    >>> f = 1/(1+fast); F = 1-f; F2 = F^2
    >>> s = 1/(1+slow); S = 1-s; S2 = S^2

    If returns are IID and WLOG ts(0) = 0 we have that (once we flip returns)
    
    >>> ts(-n) = ts(0) + rtn(0) + rtn(-1) + ... rtn(-(n-1))
    >>> fast_ewma(0) = f * ts(0) + f * F ts(-1) + f * F^n ts(-n)
    >>> = f *      (ts0)
    >>> + f * F    (ts(0) + rtn(0))
    >>> + f * F^2  (ts(0) + rtn(0) + rtn(-1))
    >>> + f * F^3  (ts(0) + rtn(0) + rtn(-1) + rtn(-2))
    >>> ...
    
    >>> fast_ewma(0) = ts(0) + F * rtn(0) + F^2 * rtn(-1) + F^3 * rtn(-2) + ...
    >>> slow_ewma(0) = ts(0) + S * rtn(0) + S^2 * rtn(-1) + S^3 * rtn(-2) + ...

    >>> crossover(0) = (F-S) rtn(0) + (F^2-S^2) * rtn(-1)...
    
    The process has zero mean and variance:
        
    >>> E(crossover^2) = \sum_{i>=1} (F^i - S^i)^2 
    >>>                = \sum_{i>=1} (F^2i + S^2i - 2 F^i * S^i)     
    >>>                = F^2 / (1 - F^2) + S^2 / (1-S^2) - 2 F*S / (1-F*S)
    

    Parameters
    ----------
    fast : int/frac
        number of days. can also be 1/(1+days) if presented as a fraction

    slow : int/frc
        number of days. can also be 1/(1+days) if presented as a fraction

    Returns
    -------
    float
        The variance of an OU process defined as ewma(dB, fast) - ewma(dB, slow) if dB is a standard Brownian Motion

    """
    f = _frac(fast); F = 1-f; F2 = F**2
    s = _frac(slow); S = 1-s; S2 = S**2
    return (F2/(1-F2) + S2/(1-S2) - 2*F*S/(1-F*S)) ** 0.5


def ewmxo_(rtn, fast, slow, vol = None, time = None, instate = None):
    """
    This is the normalized crossover function

    >>> res = (ewma(rtn, fast) - ewma(rtn, slow)) / (ewmstd(rtn, vol) * ou_factor(fast, slow))
    
    The OU factor normalizes the result so that rms(res) is approximately 1

    Parameters
    ----------
    rtn:  timeseries
        The returns of a financial process
    
    fast : int/frac
        number of days. can also be 1/(1+days) if presented as a fraction

    slow : int/frc
        number of days. can also be 1/(1+days) if presented as a fraction

    vol: int/frc
        number of days. used for calculating the volatility horizon
        
    
    :Example:
    ---------
    >>> import numpy as np; import pandas as pd; from pyg import * 
    >>> rtn = pd.Series(np.random.normal(0,1,10000),drange(-9999,0))
    >>> fast = 64; slow = 192; vol = 32; instate = None
    
    """
    state = Dict(fast = {}, slow = {}, vol = {}, cumsum = {}) if instate is None else instate
    ts = cumsum_(rtn, instate = state.get('cumsum'))
    fast_ewma_ = ewma_(ts.data, fast, time = time, instate = state.get('fast'))
    slow_ewma_ = ewma_(ts.data, slow, time = time, instate = state.get('slow'))
    vol_ = ewmstd_(rtn, vol, time = time, instate = state.get('vol')) if is_num(vol) else vol
    signal = sub_(fast_ewma_.data, slow_ewma_.data)
    normalized = div_(signal, v2na(vol_.data) * ou_factor(fast, slow))
    return Dict(data = normalized, state = Dict(fast = fast_ewma_.state, 
                                                slow = slow_ewma_.state, 
                                                vol = vol_.state))

ewmxo_.output = ['data', 'state']


def ewmacd_(ts, fast, slow, vol = None, time = None, instate = None, rms = True):
    """
    This is the normalized crossover function

    >>> res = (ewma(rtn, fast) - ewma(rtn, slow)) / (ewmstd(rtn, vol) * ou_factor(fast, slow))
    
    The OU factor normalizes the result so that rms(res) is approximately 1

    Parameters
    ----------
    rtn:  timeseries
        The returns of a financial process
    
    fast : int/frac
        number of days. can also be 1/(1+days) if presented as a fraction

    slow : int/frc
        number of days. can also be 1/(1+days) if presented as a fraction

    vol: int/frc
        number of days. used for calculating the volatility horizon
        
    
    :Example:
    ---------
    >>> import numpy as np; import pandas as pd; from pyg import * 
    >>> rtn = pd.Series(np.random.normal(0,1,10000),drange(-9999,0))
    >>> fast = 64; slow = 192; vol = 32; instate = None
    
    """
    state = Dict(fast = {}, slow = {}, vol = {}, diff = {}) if instate is None else instate
    fast_ewma_ = ewma_(ts, fast, time = time, instate = state.get('fast'))
    slow_ewma_ = ewma_(ts, slow, time = time, instate = state.get('slow'))
    rtn_ = diff_(ts, 1, time = time, instate = state.get('diff'))
    vol_ = (ewmrms_ if rms else ewmstd_)(rtn_.data, vol, time = time, instate = state.get('vol')) if is_num(vol) else vol
    signal = fast_ewma_.data - slow_ewma_.data
    normalized = div_(signal, v2na(vol_.data) * ou_factor(fast, slow))
    return Dict(data = normalized, state = Dict(fast = fast_ewma_.state, 
                                                slow = slow_ewma_.state, 
                                                vol = vol_.state,
                                                diff = rtn_.state))

ewmacd_.output = ['data', 'state']


def ewmvol_(a, n, time = None, instate = None, rms = True):
    """
    Just like ewmstd / ewmrms but calculated of prices rather than returns.
    It handles time better by delegating the time calculating for the diff as well
    
    :Example:
    ---------
    >>> import numpy as np; import pandas as pd; from pyg import * 
    >>> rtn = pd.Series(np.random.normal(0,1,10000),drange(-9999,0))
    """
    state = Dict(vol = {}, diff = {}) if instate is None else instate
    rtn_ = diff_(a, 1, time = time, instate = state.get('diff'))
    vol_ = (ewmrms_ if rms else ewmstd_)(rtn_.data, n, time = time, instate = state.get('vol'))
    return Dict(data = vol_.data, state = Dict(vol = vol_.state, 
                                                diff = rtn_.state))

ewmvol_.output = ['data', 'state']


def ewmvol(a, n, time  = None, state = None, rms = True):
    """
    calculate ewmstd/ewmrms based of a price

    :Example:
    ---------
    >>> import numpy as np; import pandas as pd; from pyg import * 
    >>> rtn = pd.Series(np.random.normal(0,1,10000),drange(-9999,0))
    >>> a = cumsum(rtn)
    >>> rtn = diff(a)
    >>> assert abs(ewmrms(rtn, 10)-ewmvol(a, 10)).max() < 1e-10
    """
    return ewmvol_(a = a, n = n, time = time, instate = state, rms = rms).data


def ewmxo(rtn, fast, slow, vol = None, time  = None, state = None):
    """
    This is the normalized crossover function

    >>> res = (ewma(rtn, fast) - ewma(rtn, slow)) / (ewmstd(rtn, vol) * ou_factor(fast, slow))
    
    The OU factor normalizes the result so that rms(res) is approximately 1

    Parameters
    ----------
    rtn:  timeseries
        The returns of a financial process
    
    fast : int/frac
        number of days. can also be 1/(1+days) if presented as a fraction

    slow : int/frc
        number of days. can also be 1/(1+days) if presented as a fraction

    vol: int/frc
        number of days. used for calculating the volatility horizon
        
    
    :Example:
    ---------
    >>> import numpy as np; import pandas as pd; from pyg import * 
    >>> rtn = pd.Series(np.random.normal(0,1,10000),drange(-9999,0))
    >>> fast = 64; slow = 192; vol = 32; instate = None    
    """
    return ewmxo_(rtn, fast, slow, vol, time = time, instate = state).data


def ewmacd(ts, fast, slow, vol = None, time  = None, state = None):
    """
    This is the normalized crossover function with price inpits

    >>> res = (ewma(rtn, fast) - ewma(rtn, slow)) / (ewmstd(rtn, vol) * ou_factor(fast, slow))
    
    The OU factor normalizes the result so that rms(res) is approximately 1

    Parameters
    ----------
    ts:  timeseries
        The price of a financial process
    
    fast : int/frac
        number of days. can also be 1/(1+days) if presented as a fraction

    slow : int/frc
        number of days. can also be 1/(1+days) if presented as a fraction

    vol: int/frc
        number of days. used for calculating the volatility horizon
        
    
    :Example:
    ---------
    >>> import numpy as np; import pandas as pd; from pyg import * 
    >>> ts = pd.Series(np.random.normal(0,1,10000),drange(-9999,0)).cumsum()
    >>> fast = 64; slow = 192; vol = 32; instate = None    
    """
    return ewmacd_(ts, fast, slow, vol, time = time, instate = state).data




import numpy as np
from pyg_timeseries._math import stdev_calculation, skew_calculation
from pyg_timeseries._decorators import compiled, first_, _data_state
from pyg_base import pd2np, loop_all

__all__ = ['cumsum', 'cumprod', ]
__all__ += ['expanding_mean', 'expanding_sum', 'expanding_rms', 'expanding_std', 'expanding_skew']

__all__ = ['cumsum_', 'cumprod_']
__all__ += ['expanding_mean_', 'expanding_sum_', 'expanding_rms_', 'expanding_std_', 'expanding_skew_']



@loop_all
@pd2np
@compiled
def _cumsum(a, t1 = 0):
    res = np.empty_like(a)
    for i in range(a.shape[0]):
        if np.isnan(a[i]):
            res[i] = np.nan
        else:
            t1 += a[i]
            res[i] = t1
    return res, t1

@loop_all
@pd2np
@compiled
def _cumprod(a, t1 = 1.):
    res = np.empty_like(a)
    for i in range(a.shape[0]):
        if np.isnan(a[i]):
            res[i] = np.nan
        else:
            t1 *= a[i]
            res[i] = t1
    return res, t1

@loop_all
@pd2np
@compiled
def _expanding_sum(a, t1 = 0.):
    res = np.empty_like(a)
    for j in range(a.shape[0]):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            t1 += a[j]
            res[j] = t1
    return res, t1

@loop_all
@pd2np
@compiled
def _expanding_mean(a, t0 = 0., t1 = 0.):
    res = np.empty_like(a)
    for j in range(a.shape[0]):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            t0 +=1.
            t1 += a[j]
            res[j] = t1/t0
    return res, t0, t1

@loop_all
@pd2np
@compiled
def _expanding_rms(a, t0 = 0., t2 = 0.):
    res = np.empty_like(a)
    for j in range(a.shape[0]):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            t0 +=1.
            t2 += a[j]**2
            res[j] = np.sqrt(t2/t0)
    return res, t0, t2

@loop_all
@pd2np
@compiled
def _expanding_std(a, t0 = 0., t1 = 0., t2 = 0.):
    res = np.empty_like(a)
    for j in range(a.shape[0]):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            t0 +=1.
            t1 += a[j]
            t2 += a[j]**2
            res[j] = stdev_calculation(t0 = t0, t1 = t1, t2 = t2)
    return res, t0, t1, t2

@loop_all
@pd2np
@compiled
def _expanding_skew(a, bias, t0 = 0., t1 = 0., t2 = 0., t3 = 0.):
    res = np.empty_like(a)
    for j in range(a.shape[0]):
        if np.isnan(a[j]):
            res[j] = np.nan
        else:
            t0 +=1
            t1 += a[j]
            t2 += a[j]**2
            t3 += a[j]**3
            res[j] = np.nan if t0<3 else skew_calculation(t0 = t0, t1 = t1, t2 = t2, t3 = t3, bias = bias, min_sample = 1)
    return res, t0, t1, t2, t3


############################################
##
## functions
##
###########################################



def cumprod(a, axis = 0, data = None, state = None):
    """
    equivalent to pandas np.exp(np.log(a).expanding().sum()).
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.
    
    data: None
        unused at the moment. Allow code such as func(live, **func_(history)) to work

    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = 1 + pd.Series(np.random.normal(0.001,0.05,10000), drange(-9999))
    >>> panda = np.exp(np.log(a).expanding().sum()); ts = cumprod(a)
    >>> assert abs(ts-panda).max() < 1e-10

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not forward fill the nans.    

    >>> a = 1 + pd.Series(np.random.normal(-0.01,0.05,100), drange(-99, 2020))
    >>> a[a<0.975] = np.nan
    >>> panda = np.exp(np.log(a).expanding().sum()); ts = cumprod(a)
    
    >>> pd.concat([panda,ts], axis=1)
    >>> 2019-09-24  1.037161  1.037161
    >>> 2019-09-25  1.050378  1.050378
    >>> 2019-09-26  1.158734  1.158734
    >>> 2019-09-27  1.158734       NaN
    >>> 2019-09-28  1.219402  1.219402
    >>>              ...       ...
    >>> 2019-12-28  4.032919  4.032919
    >>> 2019-12-29  4.032919       NaN
    >>> 2019-12-30  4.180120  4.180120
    >>> 2019-12-31  4.180120       NaN
    >>> 2020-01-01  4.244261  4.244261
    
    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> old = a.iloc[:50]        
    >>> new = a.iloc[50:]    
    >>> ts = cumprod(a)
    >>> old_ts = cumprod_(old)
    >>> new_ts = cumprod(new, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[50:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(cumprod(dict(x = a, y = a**2)), dict(x = cumprod(a), y = cumprod(a**2)))
    >>> assert eq(cumprod([a,a**2]), [cumprod(a), cumprod(a**2)])

    """
    state = state or {}
    return first_(_cumprod(a, axis = axis, **state))

def cumprod_(a, axis = 0, data = None, instate = None):
    """
    Equivalent to cumprod(a) but returns also the state variable. 
    For full documentation, look at cumprod.__doc__
    """
    state = instate or {}
    return _data_state(['data','t1'], _cumprod(a, axis = axis, **state))

cumprod_.output = ['data', 'state']



############# expanding

def expanding_mean(a, axis = 0, data = None, state = None):
    """
    equivalent to pandas a.expanding().mean().
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.
    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = a.expanding().mean(); ts = expanding_mean(a)
    >>> assert eq(ts,panda)    

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not forward fill the nans.    

    >>> a[a<0.1] = np.nan
    >>> panda = a.expanding().mean(); ts = expanding_mean(a)
    
    >>> pd.concat([panda,ts], axis=1)
    >>>                    0         1
    >>> 1993-09-23  1.562960  1.562960
    >>> 1993-09-24  0.908910  0.908910
    >>> 1993-09-25  0.846817  0.846817
    >>> 1993-09-26  0.821423  0.821423
    >>> 1993-09-27  0.821423       NaN
    >>>              ...       ...
    >>> 2021-02-03  0.870358  0.870358
    >>> 2021-02-04  0.870358       NaN
    >>> 2021-02-05  0.870358       NaN
    >>> 2021-02-06  0.870358       NaN
    >>> 2021-02-07  0.870353  0.870353
    
    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.
    
    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = expanding_mean(a)
    >>> old_ts = expanding_mean_(old)
    >>> new_ts = expanding_mean(new, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(expanding_mean(dict(x = a, y = a**2)), dict(x = expanding_mean(a), y = expanding_mean(a**2)))
    >>> assert eq(expanding_mean([a,a**2]), [expanding_mean(a), expanding_mean(a**2)])

    """
    state = state or {}
    return first_(_expanding_mean(a, axis = axis, **state))

def expanding_rms(a, axis = 0, data = None, state = None):
    """
    equivalent to pandas (a**2).expanding().mean()**0.5).
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame, list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.
    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        

    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = (a**2).expanding().mean()**0.5; ts = expanding_rms(a)
    >>> assert eq(ts,panda)    

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not forward fill the nans.    

    >>> a[a<0.1] = np.nan
    >>> panda = (a**2).expanding().mean()**0.5; ts = expanding_rms(a)
    
    >>> pd.concat([panda,ts], axis=1)
    >>>                    0         1
    >>> 1993-09-23  0.160462  0.160462
    >>> 1993-09-24  0.160462       NaN
    >>> 1993-09-25  0.160462       NaN
    >>> 1993-09-26  0.160462       NaN
    >>> 1993-09-27  0.160462       NaN
    >>>                  ...       ...
    >>> 2021-02-03  1.040346  1.040346
    >>> 2021-02-04  1.040346       NaN
    >>> 2021-02-05  1.040338  1.040338
    >>> 2021-02-06  1.040337  1.040337
    >>> 2021-02-07  1.040473  1.040473
        
    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = expanding_rms(a)
    >>> old_ts = expanding_rms_(old)
    >>> new_ts = expanding_rms(new, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])
    
    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(expanding_rms(dict(x = a, y = a**2)), dict(x = expanding_rms(a), y = expanding_rms(a**2)))
    >>> assert eq(expanding_rms([a,a**2]), [expanding_rms(a), expanding_rms(a**2)])

    """
    state = state or {}
    return first_(_expanding_rms(a, axis = axis, **state))

def expanding_sum(a, axis = 0, data = None, state = None):
    """
    equivalent to pandas a.expanding().sum().
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.
    
    data: None
        unused at the moment. Allow code such as func(live, **func_(history)) to work

    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = a.expanding().sum(); ts = expanding_sum(a)
    >>> assert eq(ts,panda)    

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not forward fill the nans.    

    >>> a[a<0.1] = np.nan
    >>> panda = a.expanding().sum(); ts = expanding_sum(a)
    
    >>> pd.concat([panda,ts], axis=1)
    >>>                    0         1
    >>> 1993-09-23          NaN          NaN
    >>> 1993-09-24          NaN          NaN
    >>> 1993-09-25     0.645944     0.645944
    >>> 1993-09-26     2.816321     2.816321
    >>> 1993-09-27     2.816321          NaN
    >>>                 ...          ...
    >>> 2021-02-03  3976.911348  3976.911348
    >>> 2021-02-04  3976.911348          NaN
    >>> 2021-02-05  3976.911348          NaN
    >>> 2021-02-06  3976.911348          NaN
    >>> 2021-02-07  3976.911348          NaN
    
    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = expanding_sum(a)
    >>> old_ts = expanding_sum_(old)
    >>> new_ts = expanding_sum(new, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(expanding_sum(dict(x = a, y = a**2)), dict(x = expanding_sum(a), y = expanding_sum(a**2)))
    >>> assert eq(expanding_sum([a,a**2]), [expanding_sum(a), expanding_sum(a**2)])

    """
    state = state or {}
    return first_(_expanding_sum(a, axis = axis, **state))

def expanding_std(a, axis = 0, data = None, state = None):
    """
    equivalent to pandas a.expanding().std().
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.
    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = a.expanding().std(); ts = expanding_std(a)
    >>> assert abs(ts-panda).max()<1e-10   

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not forward fill the nans.    

    >>> a[a<0.1] = np.nan
    >>> panda = a.expanding().std(); ts = expanding_std(a)
    
    >>> pd.concat([panda,ts], axis=1)
    >>>                    0         1
    >>> 1993-09-23       NaN       NaN
    >>> 1993-09-24       NaN       NaN
    >>> 1993-09-25       NaN       NaN
    >>> 1993-09-26       NaN       NaN
    >>> 1993-09-27       NaN       NaN
    >>>              ...       ...
    >>> 2021-02-03  0.590448  0.590448
    >>> 2021-02-04  0.590448       NaN
    >>> 2021-02-05  0.590475  0.590475
    >>> 2021-02-06  0.590475       NaN
    >>> 2021-02-07  0.590411  0.590411

    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = expanding_std(a)
    >>> old_ts = expanding_std_(old)
    >>> new_ts = expanding_std(new, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(expanding_std(dict(x = a, y = a**2)), dict(x = expanding_std(a), y = expanding_std(a**2)))
    >>> assert eq(expanding_std([a,a**2]), [expanding_std(a), expanding_std(a**2)])

    """
    state = state or {}
    return first_(_expanding_std(a, axis = axis, **state))

def expanding_skew(a, bias = False, axis = 0, data = None, state = None):
    """
    equivalent to pandas a.expanding().skew() which doesn't exist
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.
    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
            
    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))

    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = expanding_skew(a)
    >>> old_ts = expanding_skew_(old)
    >>> new_ts = expanding_skew(new, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(expanding_skew(dict(x = a, y = a**2)), dict(x = expanding_skew(a), y = expanding_skew(a**2)))
    >>> assert eq(expanding_skew([a,a**2]), [expanding_skew(a), expanding_skew(a**2)])
    """
    state = state or {}
    return first_(_expanding_skew(a, bias = bias, axis = axis, **state))


def expanding_mean_(a, axis = 0, data = None, instate = None):
    """
    Equivalent to expanding_mean(a) but returns also the state variables. 
    For full documentation, look at expanding_mean.__doc__
    """
    state = instate or {}
    return _data_state(['data','t0','t1'],_expanding_mean(a, axis = axis, **state))

expanding_mean_.output = ['data','state']

def expanding_rms_(a, axis = 0, data = None, instate = None):
    """
    Equivalent to expanding_rms(a) but returns also the state variables.
    For full documentation, look at expanding_rms.__doc__
    """
    state = instate or {}
    return _data_state(['data','t0','t2'],_expanding_rms(a, axis = axis, **state))

expanding_rms_.output = ['data','state']

def expanding_sum_(a, axis = 0, data = None, instate = None):
    """
    Equivalent to expanding_sum(a) but returns also the state variables.
    For full documentation, look at expanding_sum.__doc__
    """
    state = instate or {}
    return _data_state(['data','t1'], _expanding_sum(a, axis = axis, **state))

expanding_sum_.output = ['data','state']

def expanding_std_(a, axis = 0, data = None, instate = None):
    """
    Equivalent to expanding_mean(a) but returns also the state variables. 
    For full documentation, look at expanding_std.__doc__
    """
    state = instate or {}
    return _data_state(['data','t0', 't1', 't2'],_expanding_std(a, axis = axis, **state))

expanding_std_.output = ['data','state']

def expanding_skew_(a, bias = False, axis = 0, data = None, instate = None):
    """
    Equivalent to expanding_mean(a) but returns also the state variables. 
    For full documentation, look at expanding_skew.__doc__
    """    
    state = instate or {}
    return _data_state(['data','t0', 't1', 't2', 't3'],_expanding_skew(a, bias = bias, axis = axis, **state))

expanding_skew_.output = ['data','state']

cumsum = expanding_sum
cumsum_ = expanding_sum_




from pyg_base import is_num, is_ts, df_concat
import pandas as pd
import numpy as np
import numba

@numba.njit
def _p(x, y, vol = 0):
    if vol == 0:
        return 1. if x<y else -1. if x>y else 0.0
    else:
        one_sided_tail = 0.5 * np.exp(-abs(y-x)/vol)
        return 1-one_sided_tail if x<y else one_sided_tail

@numba.njit
def _xrank(a, w, b, vol, scale = 0 , reweight = False):
    """
    
    performs a cross-sectional rank
    
    a = np.random.normal(0,1,20)
    a[np.random.normal(0,1,20) > 1] = np.nan
    w = np.full(20, 1.)
    b = np.full(20, 1.)
    scale = 0; vol = -1; reweight = False
    a
    _xrank(a, w, b, vol)
    
    ranks a from -1 to +1 such that:
        
    i)  a[i] < a[j] ==> rank[i] < rank[j]
    ii)  rank[i] in (-1, 1)
    iii) \sum w[i] rank[i] = 0  
    
    Parameters
    ----------
    a : TYPE
        DESCRIPTION.
    w : TYPE
        DESCRIPTION.
    b : TYPE
        DESCRIPTION.
    vol : TYPE
        DESCRIPTION.
    scale : TYPE, optional
        DESCRIPTION. The default is 0.
    reweight : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    None.

    """
    not_ok = np.isnan(a)
    ok = ~not_ok
    if np.max(not_ok):
        a = a.copy(); w = w.copy(); b = b.copy()
        a[not_ok] = 0.0
        b[not_ok] = 0.0
        w[not_ok] = 0.0
    wb = w * b
    total_wb = np.sum(wb)
    if total_wb == 0:
        return np.full_like(a, np.nan)
    else:
        r = np.zeros_like(a)
        wb = wb / total_wb

    if vol < 0:    
        wba = wb * a
        m1 = np.sum(wba)
        m2 = np.sum(wba * a)
        vol = (m2 - m1**2) ** 0.5

    for i in range(a.shape[0]):
        if ok[i] and w[i]!=0:
            for j in range(i):
                if ok[j] and w[j]!=0:
                    qq = _p(a[i], a[j], vol)
                    pp = 1-qq
                    r[i] += (2*pp-1) * wb[j]
                    r[j] += (2*qq-1) * wb[i]
    
    if scale == 0:
        std = 1
    elif scale == 1: # scale weightes so that total weight = 1
        total_w = np.sum(w)
        w = w / total_w
        std = np.sum((w*r)**2*(1-b**2)) ** 0.5
        r = r/std
    elif scale == 2:
        std = (np.sum(r**2) - np.sum(r)**2) ** 0.5
        r = r/std
    elif scale == 3:
        total_w = np.sum(w)
        w = w / total_w
        std = np.sum(w*(r**2)) ** 0.5
        r = r/std
    
    r[not_ok] = np.nan
    if reweight:
        r = r * w
    return r

@numba.njit
def _xrank_2d(a, w, b, vol, scale, reweight):
    res = np.empty_like(a)
    for i in range(a.shape[0]):
        res[i] = _xrank(a = a[i], w = w[i], b = b[i], vol = vol[i], scale = scale , reweight = reweight)
    return res

def xrank(a, weight = None, beta = None, vol = True, scale = 0 , reweight = False, columns = None):
    """
    
    performs a cross-sectional rank
    
    a = np.random.normal(0,1,20)
    a[np.random.normal(0,1,20) > 1] = np.nan
    w = np.full(20, 1.)
    b = np.full(20, 1.)
    scale = 0; vol = -1; reweight = False
    a
    _xrank(a, w, b, vol)
    
    ranks a from -1 to +1 such that:
        
    i)  a[i] < a[j] ==> rank[i] < rank[j]
    ii)  rank[i] in (-1, 1)
    iii) \sum w[i] rank[i] = 0  
    
    Parameters
    ----------
    a : TYPE
        DESCRIPTION.
    w : TYPE
        DESCRIPTION.
    b : TYPE
        DESCRIPTION.
    vol : TYPE
        DESCRIPTION.
    scale : TYPE, optional
        DESCRIPTION. The default is 0.
    reweight : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    None.
    
    :Example:
    ---------
    >>> a = pd.DataFrame(np.random.normal(0,1,(1000,20)), drange(-999))
    >>> aa = cumsum(a)
    >>> aa.plot()
    >>> beta = weight = None
    >>> vol = True; scale = 0; columns = None
    >>> res = xrank(aa)
    >>> res.plot()

    """
    a = df_concat(a, columns).ffill()
    index = a.index
    cols = a.columns
    a_ = a.values
    if weight is None:
        w = np.full_like(a_, 1.)
    elif is_num(weight):
        w = np.full_like(a_, weight)
    else:
        w = df_concat(weight, columns).reindex(index, method = 'ffill')
    if beta is None:
        b = np.full_like(a_, 1.)
    elif is_num(beta):
        b = np.full_like(a_, beta)
    else:
        b = df_concat(beta, columns).reindex(index, method = 'ffill')
    if vol is True:
        vol = -1
    if is_ts(vol):
        vol - vol.reindex(index, method = 'ffill')
        if isinstance(vol, pd.DataFrame) and vol.shape[1] == 1:
            vol = vol.iloc[:,0]
    else:
        vol = np.full(a_.shape[0], vol)        
        
    b, w, vol = [df.values if is_ts(df) else df for df in (b,w,vol)]
    res = _xrank_2d(a_, w, b, vol, scale, reweight)
    return pd.DataFrame(res, index, cols)


import numpy as np
from pyg_timeseries._decorators import compiled


@compiled
def stdev_calculation(t0, t1, t2, default = np.nan):
    if t0 > 1:
        p = t0-1
        return np.sqrt(t2/p - (t1**2)/(p*t0))
    else:
        return default

@compiled
def std_calculation(t0, t1, t2, default = np.nan):
    if t0 > 1:
        p = t0 #biased :-)
        return np.sqrt(t2/p - (t1**2)/(p*t0))
    else:
        return default


@compiled
def stdev_calculation_ewm(t0, t1, t2, w2, min_sample, bias = False):
    """
    The nicest calculation of unbiased variance under variable weights is available here:
    https://mathoverflow.net/questions/11803/unbiased-estimate-of-the-variance-of-a-weighted-mean
    """
    if t0<=min_sample:
        return np.nan
    variance = t2/t0 - (t1/t0)**2
    if variance<0:
        return np.nan
    elif bias:
        return np.sqrt(variance)
    else:
        r = 1 - w2/(t0**2)
        return np.sqrt(variance/r)

@compiled
def variance_calculation_ewm(t0, t1, t2, w2, min_sample, bias = False):
    """
    The nicest calculation of unbiased variance under variable weights is available here:
    https://mathoverflow.net/questions/11803/unbiased-estimate-of-the-variance-of-a-weighted-mean
    """
    if t0<=min_sample:
        return np.nan
    variance = t2/t0 - (t1/t0)**2
    if variance<0:
        return np.nan
    elif bias:
        return variance
    else:
        r = 1 - w2/(t0**2)
        return variance/r


@compiled
def cor_calculation(t0, a1, a2, b1, b2, ab, min_sample):
    if t0<=min_sample:
        return np.nan
    Eab = ab/t0
    Ea = a1/t0
    Eb = b1/t0
    STDa = std_calculation(t0, a1, a2) # to match the pandas implementation, we use the biased std_calculation estimator
    STDb = std_calculation(t0, b1, b2)
    denom = STDa * STDb
    if denom > 0:
        return (Eab - Ea*Eb)/denom
    else:
        return np.nan


@compiled
def cor_calculation_ewm(t0, a1, a2, b1, b2, w2, ab, min_sample, bias = False):
    if t0<=min_sample:
        return np.nan
    Eab = ab/t0
    Ea = a1/t0
    Eb = b1/t0
    STDa = stdev_calculation_ewm(t0, a1, a2, w2, min_sample = min_sample, bias = bias)
    STDb = stdev_calculation_ewm(t0, b1, b2, w2, min_sample = min_sample, bias = bias)
    denom = STDa * STDb
    if denom > 0:
        return (Eab - Ea*Eb)/denom
    else:
        return np.nan

@compiled
def corr_calculation_ewm(ab0, a0, a1, a2, aw2, b0, b1, b2, bw2, ab, min_sample, bias = False):
    if ab0<=min_sample or a0<=min_sample or b0<=min_sample:
        return np.nan
    Eab = ab/ab0
    Ea = a1/a0
    Eb = b1/b0
    STDa = stdev_calculation_ewm(a0, a1, a2, aw2, min_sample = min_sample, bias = bias)
    STDb = stdev_calculation_ewm(b0, b1, b2, bw2, min_sample = min_sample, bias = bias)
    denom = STDa * STDb
    if denom > 0:
        return (Eab - Ea*Eb)/denom
    else:
        return np.nan



@compiled
def LR_calculation_ewm(t0, a1, a2, b1, b2, w2, ab, min_sample, bias = False):
    if t0<=min_sample:
        return np.nan, np.nan
    Eab = ab/t0
    Ea = a1/t0
    Eb = b1/t0
    VARa = variance_calculation_ewm(t0, a1, a2, w2, min_sample = min_sample, bias = bias)
    if VARa > 0:
        covar = Eab - Ea*Eb
        m = covar/VARa
        c = Eb - m * Ea
        return c, m
    else:
        return np.nan, np.nan
    
@compiled
def skew_calculation(t0, t1, t2, t3, bias, min_sample):
    if t0<=max(min_sample,2):
        return np.nan
    t0 = t0 * 1.
    m1 = t1 / t0
    m2 = t2 / t0 - m1 ** 2
    if m2 > 0:
        m3 = (t3/t0 - 3 * m1 * t2 / t0 + 2 * (m1**3))
        biased = m3 / (m2**1.5)
        if bias:
            return biased
        else:
            return biased  * np.sqrt(t0 * (t0-1)) / (t0-2) 
    else:
        return np.nan

